/************************************************/
/*						*/
/*      Copyright:				*/
/*	 Jean-Marc Pigeon <jmp@safe.ca>	 2009	*/
/*						*/
/************************************************/
/* This program is free software; you can 	*/
/* redistribute it and/or modify it under the 	*/
/* terms of the GNU General Public License as	*/
/* published by the Free Software Foundation	*/
/* version 2 of the License			*/
/*						*/
/* This program is distributed in the hope that */
/* it will be useful, but WITHOUT ANY WARRANTY; */
/* without even the implied warranty of		*/
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR	*/
/* PURPOSE.  See the GNU General Public License	*/
/* for more details.				*/
/*						*/
/* You should have received a copy of the GNU	*/
/* General Public License along with this 	*/
/* program; if not, write to the Free Software	*/
/* Foundation, Inc., 51 Franklin Street,	*/
/* Fifth Floor, Boston, MA  02110-1301, USA.	*/
/************************************************/
/*						*/
/*	UNICNT:					*/
/*						*/
/************************************************/
#include        <sys/mount.h>
#include        <sys/syscall.h>
#include        <sys/time.h>
#include        <sys/wait.h>
#include        <sys/stat.h>
#include	<time.h>
#include	<sched.h>
#include	<signal.h>
#include	<stdio.h>
#include	<string.h>
#include	<syslog.h>
#include	<errno.h>
#include	<fcntl.h>
#include	<unistd.h>

#include	"lowtyp.h"
#include	"dbgmem.h"
#include	"subapl.h"
#include	"unicnt.h"

/*container filesystem				*/
#define	HOMEFS	"HOMEFS"	/*container HOME*/
#define	CLONPID	"first.pid"	/*cont process 1*/
#define	MTAB	"/etc/mtab"	/*mtab file	*/

#define	DDEV	"/dev"		/*system /dev	*/
#define	DEVFS	"devtmpfs"	/*/dev systype	*/

#define	ROOTFS	"rootfs"	//Container main root directory
				//
/*systemd sugget to have /proc/ and /sys to be	*/
/*mounted read only (MS_RDONLY).		*/
#define	MNTMODE	0		/*Mounting mode	*/

/*

*/
/************************************************/
/*						*/
/*	Defining pivot_root system call		*/
/*						*/
/************************************************/
static int pivot_root(const char * new_root,const char * put_old)

{
#ifndef __NR_pivot_root
#pragma message ("the pivot_root syscall is not available within \"sys/syscall.h\"")
#pragma message ("-> The pivot_root system call will generate an alert")
(void) apl_alert(0,"%s need the pivot_root system call!!!.",appname);
errno=ENOSYS;
return -1;
#else
return(syscall(__NR_pivot_root, new_root, put_old));
#endif
}
/*

*/
/************************************************/
/*						*/
/*	Returning the container path.		*/
/*	contpath array is dynamically assigned	*/
/*	must be freed when not needed anymore.	*/
/*						*/
/************************************************/
static char *getcontpath(char *contname)

{
char *homefs;
char *contpath;
char *vzpath;


homefs=getenv(HOMEFS);
contpath=(char *)0;
if (homefs==(char *)0) {
  vzpath=apl_appdir(d_vzgot);
  (void) asprintf(&contpath,"%s/%s",vzpath,contname);
  (void) free(vzpath);
  }
else
  (void) asprintf(&contpath,"%s",homefs);
return contpath;
}
/*

*/
/************************************************/
/*						*/
/*	Checking and Waiting to be ready to	*/
/*	accept user mapping.			*/
/*	Return true if mapping possible, or	*/
/*	false if trouble or timeout.		*/
/*						*/
/************************************************/
static _Bool iscloneready(pid_t clonepid,char *map)

{
#define OPEP    "unicnt.c:iscloneready"
#define	SLICE	10000	//10 Millisecond
#define MXSEC	5       //maximun seconds waiting time
			//for CLONE_NEWUSER to be actif
			//within clone process
#define	MXWAIT	((MXSEC*1000000)/SLICE)

_Bool done;

done=false;
for (int i=0;i<MXWAIT;i++) {
  int handle;
  int status;
  int errloc;
  char mode[200];

  (void) usleep(SLICE);
  (void) memset(mode,'\000',sizeof(mode));
  if ((handle=open(map,O_RDONLY))<0) {
    (void) apl_alert(0,"%s Unable to open file <%s> (error=<%s>)",
                        OPEP,map,strerror(errno));
    break;      //no need to try anymore!
    }
  status=read(handle,mode,sizeof(mode)-1);
  errloc=errno;
  (void) close(handle);
  if (status<0) {
    (void) apl_alert(0,"%s Unable to read file <%s> (error=<%s>)",
                        OPEP,map,strerror(errloc));
    break;      //no need to try anymore!
    }
  if (strlen(mode)==0) {
    (void) apl_alert(2,"%s clone process '%d' detected in CLONE_NEWUSER mode",
                        OPEP,clonepid);
    done=true;
    break;      //check successful
    }
  if ((i%(MXWAIT/MXSEC))==0) {
    (void) apl_alert(0,"%s Waiting (%2d/%2d) for process '%d' to "
                       "be in CLONE_NEWUSER mode",OPEP,i,MXWAIT,clonepid);
    }
  if (kill(clonepid,0)<0) {
    (void) apl_alert(0,"%s clone process'%d' Premature exit (bug?)",
		    	OPEP,clonepid);
    break;
    }
  }
return done;

#undef	MXSEC
#undef  SLICE
#undef  MXWAIT
#undef  OPEP
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check the current uid and	*/
/*	gid to be "right"			*/
/*	Wait at max for delay second.		*/
/*						*/
/************************************************/
_Bool cnt_wait_goodid(int delay,uid_t uid,gid_t gid)

{
#define	SLICE	10000	//waiting time between check

_Bool done;

done=false;
delay*=(1000000/SLICE);
for (int i=0;i<delay;i++) {
  (void) usleep(SLICE);
  if ((getuid()==uid)&&(getgid()==gid)) {
    done=true;
    break;
    }
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to set the container system	*/
/*	directory				*/
/*						*/
/************************************************/
static _Bool do_sysmount(char *rootfs)

{
#define OPEP    "vzgot:do_sysmount"

_Bool done;
char ppath[50];
int phase;
_Bool proceed;

done=false;
phase=0;
proceed=true;
while (proceed==true) {
  (void) apl_alert(2,"%s phase='%d'",OPEP,phase);
  switch (phase) {
    case 0      :       //Empty
      break;
    case 1      :       //Empty
      (void) snprintf(ppath,sizeof(ppath),"%s/%s",rootfs,"proc");
      if (mount("proc",ppath,"proc",MS_NODEV|MS_NOSUID|MS_NOEXEC,"")<0)  {
        (void) apl_alert(0,"%s Unable to mount <%s> within container "
                           "(error=<%s>)",OPEP,ppath,strerror(errno));
        phase=999;          //trouble trouble
        }
      break;
    case 2      :       // /proc/sys/net
      (void) snprintf(ppath,sizeof(ppath),"%s/%s",rootfs,"proc/sys/net");
      if (mount("none",ppath,"proc",MS_NODEV|MS_NOSUID|MS_NOEXEC,"")<0)  {
        (void) apl_alert(0,"%s Unable to mount <%s> within container "
                           "(error=<%s>)",OPEP,ppath,strerror(errno));
        phase=999;          //trouble trouble
        }
      break;
    case 3      :       //mounting /proc/sys
      (void) snprintf(ppath,sizeof(ppath),"%s/%s",rootfs,"proc/sys");
      if (mount("proc",ppath,"proc",MS_RDONLY,"")<0)  {
        (void) apl_alert(0,"%s Unable to mount <%s> within container "
                           "(error=<%s>)",OPEP,ppath,strerror(errno));
        phase=999;          //trouble trouble
        }
      break;
    case 4      :       //mounting sys
      (void) snprintf(ppath,sizeof(ppath),"%s/%s",rootfs,"sys");
      if (mount("sysfs",ppath,"sysfs",MS_NODEV|MS_NOSUID|MS_NOEXEC|MS_RDONLY,"")<0)  {
        (void) apl_alert(0,"%s Unable to mount <%s> within container "
                           "(error=<%s>)",OPEP,ppath,strerror(errno));
        phase=999;          //trouble trouble
        }
      break;
    case 5      :       //Empty
      break;
    case 6      :       //Empty
      break;
    case 7      :       //Empty
      done=true;
      break;
    default     :       //SAFE Guard
      proceed=false;
      break;
    }
  phase++;
  }
return done;

#undef  OPEP
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to set the container working	*/
/*	"root" user at kernel level, used by	*/
/*	container withn CLONE_NEWUSER mode.	*/
/*						*/
/************************************************/
_Bool cnt_mapcontids(pid_t clonepid,uid_t contuid,gid_t contgid)

{
#define OPEP    "unicnt.c:cnt_mapcontids"
#define MXID    4294967295		//'Standard' max uid
#define MXMAP   sizeof(names)/sizeof(char *)

static char *names[]={
        "uid_map",
        "setgroups",
        "gid_map"
        };


_Bool done;
char locnames[MXMAP][100];
char data[MXMAP][100];
int phase;
_Bool proceed;

done=false;
(void) memset(locnames,'\000',sizeof(locnames));
(void) memset(data,'\000',sizeof(data));
phase=0;
proceed=true;
while (proceed==true) {
  (void) apl_alert(3,"%s phase='%d'",OPEP,phase);
  switch (phase) {
    case 0        :     //preparing file name
      for (int i=0;i<MXMAP;i++) 
        (void) snprintf(locnames[i],sizeof(locnames[i]),
                         "/proc/%d/%s",clonepid,names[i]);
      break;
    case 1        :     //waiting for clone process to be in CLONE_NEWUSER 
      if (iscloneready(clonepid,locnames[0])==false) {
        (void) apl_alert(0,"%s Unable to map user in cloned "
                         "process='%d' (aborting!)",OPEP,clonepid);
        if (kill(clonepid,0)==0) {
          (void) apl_alert(0,"%s terminating unresponsive clone process (pid=%d)",
                            OPEP,clonepid);
          (void) kill(clonepid,SIGTERM);
          (void) sleep(2);
          if (kill(clonepid,0)==0) {
            (void) apl_alert(0,"%s Overkill unresponsive clone process (pid=%d)",
                            OPEP,clonepid);
            (void) kill(clonepid,SIGKILL);
            }
          }
        phase=999;      //no need to go further
        }
      break;
    case 2        :     //preparing data
      (void) snprintf(data[0],sizeof(data[0]),"0 %d %ld",contuid,MXID);
      (void) snprintf(data[1],sizeof(data[1]),"allow");
      (void) snprintf(data[2],sizeof(data[2]),"0 %d %ld",contgid,MXID);
      break;
    case 3        :     //storing data in process map
      for (int i=0;i<MXMAP;i++) {
        int handle;

        if ((handle=open(locnames[i],O_WRONLY))<0) {
	  (void) apl_alert(0,"%s, unable to open <%s> (error=<%s>)",
                            OPEP,locnames[i],strerror(errno));
          phase=999;
          break;        //Trouble no need to go further
          }
        if (write(handle,data[i],strlen(data[i]))<0) {
	  (void) apl_alert(0,"%s, unable to write <%s> to <%s> (error=<%s>)",
                            OPEP,data[i],locnames[i],strerror(errno));
          phase=999;    //no good status
          }
        (void) close(handle);
        }
      break;
    case 4      :       //everyhing is right, lets report it
      done=true;
      break;
    default     :       //SAFE guard
      proceed=false;
      break;
    }
  phase++;
  }
return done;

#undef  MXMAP
#undef  OPEP
}
/*

*/
/************************************************/
/*						*/
/*	Procedure used by container (once	*/
/*	cloned) to make sure the current ID is	*/
/*	propperly set.				*/
/*						*/
/************************************************/
_Bool cnt_wait_for_mapuser(uid_t contuid,gid_t contgid)

{
#define OPEP    "unicnt.c:cnt_wait_for_mapuse"
#define	SLICE	10000	//10 Milli-seconde
#define MXSEC	2       //maximun secondes waiting time
			//for CLONE_NEWUSER to be actif
			//within clone process
#define	MXWAIT	((MXSEC*1000000)/SLICE)

_Bool done;

done=false;
for (int i=0;i<MXWAIT;i++) {
  (void) usleep(1000);
  if ((getuid()==contuid)&&(getgid()==contgid)) {
    done=true;
    break;
    }
  }
return done;

#undef	MXSEC
#undef  SLICE
#undef  MXWAIT
#undef  OPEP
}
/*

*/
/************************************************/
/*						*/
/*	procedure to pivot the container root	*/
/*	file system.				*/
/*						*/
/************************************************/
_Bool cnt_pivot(char *contname,int cloneflgs)

{
#define	OPEP	"unicnt.c:cnt_pivot,"
#define	HOSTFS	"hostfs_root"	//container host PIVOT

int done;
char *rootfs;
char *contpath;
char ppath[200];
int phase;
int proceed;

done=false;
rootfs=(char *)0;
contpath=getcontpath(contname);
(void) asprintf(&rootfs,"%s/%s",contpath,ROOTFS);
phase=0;
proceed=true;
while (proceed==true) {
  (void) apl_alert(2,"%s phase='%d'",OPEP,phase);
  switch (phase) {
    case 0	:	//Ensure that 'new_root' and its parent mount don't have
			//hared propagation (which would cause pivot_root() to
              		//return an error), and prevent propagation of mount
			//events to the initial mount namespace.
      if (mount(NULL,"/",0,MS_REC|MS_SLAVE,NULL)<0) {
        (void) apl_alert(0,"%s unable to stop shared propagation in container "
			   "(error=%s)",OPEP,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	//Ensure that 'rootfs' is a mount point.
      if (mount(rootfs,rootfs,NULL,MS_BIND,NULL)<0) {
        (void) apl_alert(0,"%s unable to bind container <%s> directory "
			   "(error=%s)",
			   OPEP,rootfs,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	//Mounting system directory
      if (do_sysmount(rootfs)==false) {
        (void) apl_alert(0,"%s unable to mount container <%s> system directory"
			    OPEP,contname);
	phase=999;	//trouble trouble
        }
      break;
    case 3	:	///Empty
      break;
    case 4	:	//creating the "pivot" repository
      (void) snprintf(ppath,sizeof(ppath),"%s/%s",rootfs,HOSTFS);
      if (mkdir(ppath,0777)<0) {
	switch (errno) {
	  case EEXIST	:	//this could append (its fine)
	    (void) apl_alert(0,"%s, pivoting <%s> directory  allready existing",
                               OPEP,ppath);
	    break;
	  default	:	//Unexpected result
	    (void) apl_alert(0,"%s Unable to create pivoting <%s> directory "
                               "(error=<%s>)",OPEP,HOSTFS,strerror(errno));
            phase=999; 
	    break;
	  }
	}
      break;
    case 5	:	/*moving to rootfs	*/
      if (pivot_root(rootfs,ppath)<0) {
        (void) apl_alert(0,"%s unable to do pivot to <%s> old root <%s> "
			   "to <%s>(error=%s)",
			   OPEP,rootfs,ppath,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 6	:	//chrooting to root directory
      if (chdir("/")<0) {
        (void) apl_alert(0,"%s unable to reach container <%s> "
			   "root directory (error=%s)",
			   OPEP,contname,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 7	:	/*checking /proc dir	*/
      break;
    case 8	:	//empty
      break;
    case 9	:	/*mounting proc dir	*/
      break;
    case 10	:	//mounting DSYS if needed
      break;
    case 11	:	//empty
      if (umount2(HOSTFS,MNT_DETACH)<0) {
        (void) apl_alert(0,"%s unable to unmount container <%s> "
			   "oldroot <%s> (error=%s)",
			   OPEP,contname,HOSTFS,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 12	:	/*mounting dev RD_ONLY?	*/
      if (rmdir(HOSTFS)<0) 
        (void) apl_alert(0,"%s unable to remove <%s> directory "
			   "from container <%s>, (error=%s)",
			   OPEP,HOSTFS,contname,strerror(errno));
      break;
    case 13	:	/* everything fine	*/
      (void) fprintf(stdout,"%s chroot done for container <%s>\n",
			    apl_ascsystime(time((time_t *)0)),contname);
      (void) fflush(stdout);
      done=true;
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(contpath);
(void) free(rootfs);
return done;


#undef	HOSTFS
#undef	DPROC
#undef	OPEP
}
/*

*/
/************************************************/
/*						*/
/*	procedure to set a file with the clone	*/
/*	ip number.				*/
/*						*/
/************************************************/
int cnt_setclonepid(char *contname,pid_t cpid)

{
int done;
FILE *fichier;
char *filename;
char *contpath;
int phase;
int proceed;

done=false;
fichier=(FILE *)0;
filename=(char *)0;
contpath=getcontpath(contname);
(void) asprintf(&filename,"%s/%s",contpath,CLONPID);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*opening the file	*/
      if ((fichier=fopen(filename,"w"))==(FILE *)0) {
	(void) apl_alert(0,"%s, Unable to open pidfile <%s> (error=<%s>)",
			    appname,filename,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*writing  PID in it	*/
      if (fprintf(fichier,"%d\n",cpid)<0) {
	(void) apl_alert(0,"%s, Unable to write pidfile in container <%s> (error=<%s>)",
			    appname,contname,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      (void) fclose(fichier);
      break;
    case 2	:	/*everything fine	*/
      done=true;
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(filename);
(void) free(contpath);
return done;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to get the clone pid stored	*/
/*	within the clonpid file			*/
/*						*/
/************************************************/
pid_t cnt_getclonepid(char *contname)

{
pid_t clonepid;
FILE *fichier;
char *filename;
char *contpath;

clonepid=(pid_t)0;
contpath=getcontpath(contname);
(void) asprintf(&filename,"%s/%s",contpath,CLONPID);
if ((fichier=fopen(filename,"r"))!=(FILE *)0) {
  char strloc[80];

  (void) fgets(strloc,sizeof(strloc)-1,fichier);
  (void) sscanf(strloc,"%d",&clonepid);
  (void) fclose(fichier);
  }
(void) free(filename);
(void) free(contpath);
return clonepid;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to remove the file with the	*/
/*	clone pid.				*/
/*						*/
/************************************************/
int cnt_rmclonepid(char *contname)

{
int done;
char *filename;
char *contpath;

done=true;
filename=(char *)0;
contpath=getcontpath(contname);
(void) asprintf(&filename,"%s/%s",contpath,CLONPID);
if (unlink(filename)<0) {
  (void) apl_alert(0,"%s, Unable to remove pidfile in container <%s> (error=<%s>)",
		     appname,contname,strerror(errno));
  done=false;
  }
(void) free(filename);
(void) free(contpath);
return done;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to initialize the HOST side	*/
/*						*/
/************************************************/
_Bool cnt_initscript(const char *scriptname,const char *fmt,...)

{
#define	OPEP	"unicnt.c:cnt_initscript"
#define	BSIZE	2048

_Bool done;
va_list args;
FILE *canal;
char *varlibpath;
char *parmlst;
char *cmd;
int phase;
int proceed;

done=false;
va_start(args,fmt);
canal=(FILE *)0;
varlibpath=apl_appdir(d_varlib);
(void) vasprintf(&parmlst,fmt,args);
(void) asprintf(&cmd,"%s/shell/%s %s",varlibpath,scriptname,parmlst);
phase=0;
proceed=true;
while (proceed==true) {
  (void) apl_alert(3,"%s Phase='%d' (done='%d')",OPEP,phase,done);
  switch (phase) {
    case 0	:	/*opening pip channel	*/
      if ((canal=popen(cmd,"r"))==(FILE *)0) {
	(void) apl_alert(0,"%s, Unable to pipe cmd <%s> (error=<%s>)",
			    OPEP,cmd,strerror(errno));
	phase=999;	/*trouble trouble	*/	
	}
      break;
    case 1	:	/*do we have feed back	*/
      if (canal!=(FILE *)0) {	/*always	*/
	char buffer[BSIZE];

	while (fgets(buffer,BSIZE,canal)!=(char *)0) {
	  (void) apl_cleanstring(buffer);
	  (void) apl_alert(2,"%s, pipe say: <%s>",appname,buffer);
	  (void) fprintf(stdout,"%s\n",buffer);
	  (void) fflush(stdout);
	  }
	}
      break;
    case 2	:	/*closing pip		*/
      if (pclose(canal)<0) {
	(void) apl_alert(0,"%s, Unable to close pipe, cmd <%s> (error=<%s>)",
			    OPEP,cmd,strerror(errno));
	phase=999;	//trouble trouble
	}
      break;
    case 3	:	//reporting everython OK
      done=true;
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(cmd);
(void) free(parmlst);
(void) free(varlibpath);
va_end(args);
return done;

#undef	OPEP
}
/*

*/
/************************************************/
/*						*/
/*	procedure to inject a command within	*/
/*	the container.				*/
/*						*/
/************************************************/
int cnt_injectcmd(const char *scriptname,char *contname,int contpid,char *params)

{
#define	OPEP	"unicnt.c:cnt_injectcmd,"
#define	FDS	3
#define	CMDSEQ	"%s/shell/%s %s %d %s"

int status;
char *varlibpath;
char *cmd;
int phase;
int proceed;

status=0;
varlibpath=apl_appdir(d_varlib);
(void) asprintf(&cmd,CMDSEQ,varlibpath,scriptname,contname,contpid,params);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	//protecting already std devices?
      break;
    case 1	:	//executing the command
      status=system(cmd);
      break;
    case 2	:	//putting back devices
      break;
    default	:	//SAFE Guard
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(cmd);
(void) free(varlibpath);
return status;
#undef OPEP
}
/*

*/
/************************************************/
/*						*/
/*	procedure to extract the container	*/
/*	architecture available in the 'arch' 	*/
/*	file.					*/
/*						*/
/************************************************/
char *cnt_getarch(char *contname)

{
static char *availarch[]={
	"i386","i686","x86_64",(char *)0
	};

char *arch;
FILE *fichier;
char *contpath;
char *filename;
int phase;
int proceed;
char buffer[200];

arch=(char *)0;
fichier=(FILE *)0;
filename=(char *)0;
contpath=getcontpath(contname);
(void) asprintf(&filename,"%s/%s",contpath,"arch");
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*let open the file	*/
      if ((fichier=fopen(filename,"r"))==(FILE *)0) {
	(void) apl_alert(0,"%s Unable to open file <%s> (error=<%s>)",
			   appname,filename,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*reading first line	*/
      if (apl_getstr(fichier,buffer,sizeof(buffer),'#')==(char *)0) {
	(void) apl_alert(0,"%s Unable to read file <%s> (file empty?)",
			   appname,filename);
	phase=999;	/*trouble trouble	*/
	}
      (void) fclose(fichier);
      break;
    case 2	:	/*scanning line		*/
      (void) apl_cleanstring(buffer);
      if (strlen(buffer)>0) { 	/*always?	*/
	int i;

        for (i=0;availarch[i]!=(char *)0;i++) {
	  if (strcmp(availarch[i],buffer)==0) {
	    arch=strdup(availarch[i]);
	    proceed=false;	/*found arch	*/
	    break;
	    }
	  }
	}
      break;
    case 3	:	/*scanning line		*/
      (void) apl_alert(0,"%s file <%s>, arch <%s> unknown, setting default",
			   appname,filename,buffer);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(filename);
(void) free(contpath);
if (arch==(char *)0) {
  arch=strdup(availarch[0]);
  }
return arch;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to extract the container	*/
/*	distribution type available in the  	*/
/*	'dist' file.				*/
/*						*/
/************************************************/
char *cnt_getdist(char *contname)

{
char *dist;
FILE *fichier;
char *contpath;
char *filename;
int phase;
int proceed;
char buffer[200];

dist=(char *)0;
fichier=(FILE *)0;
filename=(char *)0;
contpath=getcontpath(contname);
(void) asprintf(&filename,"%s/%s",contpath,"dist");
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*let open the file	*/
      if ((fichier=fopen(filename,"r"))==(FILE *)0) {
	(void) apl_alert(0,"%s Unable to open file <%s> (error=<%s>)",
			   appname,filename,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*reading first line	*/
      if (apl_getstr(fichier,buffer,sizeof(buffer),'#')==(char *)0) {
	(void) apl_alert(0,"%s Unable to read file <%s> (file empty?)",
			   appname,filename);
	phase=999;	/*trouble trouble	*/
	}
      (void) fclose(fichier);
      break;
    case 2	:	/*scanning line		*/
      (void) apl_cleanstring(buffer);
      dist=strdup(buffer);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(filename);
(void) free(contpath);
if (dist==(char *)0) {
  dist=strdup("unknown");
  }
return dist;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to assign a new STDOUT/STDERR	*/
/*	and close STDIN. Done ONLY if the	*/
/*	process is in background mode.		*/
/*						*/
/************************************************/
void cnt_setstdio(int foreground,char *contname,char *outname)

{
#define	OPEP	"unicnt.c:cnt_setstdio,"

if (foreground==false) {
  int newstdout;
  int newstderr;
  int phase;
  int proceed;
  char *contpath;
  char *filename;

  newstdout=-1;
  newstderr=-1;
  phase=0;
  proceed=true;
  contpath=getcontpath(contname);
  (void) asprintf(&filename,"%s/%s.stdout",contpath,outname);
  while (proceed==true) {
    switch (phase) {
      case 0	:	/*open the new stdout	*/
	(void) unlink(filename);
	if ((newstdout=open(filename,O_CREAT|O_RDWR|O_TRUNC,S_IRUSR|S_IWUSR|S_IRGRP))<0) {
	  (void) apl_alert(0,"%s, Unable to open <%s> (error=<%s>)",
			      OPEP,filename,strerror(errno));
	  phase=999;	/*trouble trouble	*/
	  }
	break;
      case 1	:	/*open the new stderr	*/
	(void) free(filename);
        (void) asprintf(&filename,"%s/%s.stderr",contpath,outname);
	(void) unlink(filename);
	if ((newstderr=open(filename,O_CREAT|O_RDWR|O_TRUNC,S_IRUSR|S_IWUSR|S_IRGRP))<0) {
	  (void) apl_alert(0,"%s, Unable to open <%s> (error=<%s>)",
			      OPEP,filename,strerror(errno));
	  (void) close(newstdout);
	  phase=999;	/*trouble trouble	*/
	  }
	break;
      case 2	:	/*duplicating stdout	*/
	if (dup2(newstdout,1)<0) {
	  (void) apl_alert(0,"%s, Unable to dup2 stdout (error=<%s>)",
			      OPEP,strerror(errno));
	  (void) close(newstderr);
	  (void) close(newstdout);
	  phase=999;
	  }
	break;
      case 3	:	/*duplicating stderr	*/
	(void) close(newstdout);
	if (dup2(newstderr,2)<0) {
	  (void) apl_alert(0,"%s, Unable to dup2 stderr (error=<%s>)",
			      OPEP,strerror(errno));
	  (void) close(newstderr);
	  phase=999;
	  }
	break;
      case 4	:	/*setting new std	*/
	if ((stdout=fdopen(1,"w"))==(FILE *)0) {
	  (void) apl_alert(0,"%s, Unable to fdopen stdout (error=<%s>)",
			      OPEP,strerror(errno));
	  }
	if ((stderr=fdopen(2,"w"))==(FILE *)0) {
	  (void) apl_alert(0,"%s, Unable to fdopen stderr (error=<%s>)",
			      OPEP,strerror(errno));
	  }
	(void) fclose(stdin);
	break;
      case 5	:	/*write time stamp	*/
	(void) fprintf(stdout,"stdout start: %s\n",apl_ascsysdatetime(time((time_t *)0)));
	(void) fprintf(stderr,"stderr start: %s\n",apl_ascsysdatetime(time((time_t *)0)));
	(void) fflush(stdout);
	(void) fflush(stderr);
	break;
      default	:	/*SAFE Guard		*/
	proceed=false;
        break;
      }
    phase++;
    }
  (void) free(filename);
  (void) free(contpath);
  }
#undef	OPEP
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to wait for entry on the	*/
/*	container console and forward it	*/
/*	on the Master_container console file.	*/
/*	A Management expiration is done		*/
/*	according time.				*/
/*	Status is 0 and 2, if process shutdown	*/
/*	Status is 1, if process need reboot	*/
/*						*/
/************************************************/
int cnt_mstconsole(char *contname,pid_t cntpid)

{
#define	OPEP	"unicnt.c:cnt_cmstonsole,"
#define	RLX	30	/*number of sec		*/

int count;
int relaxtime;
int status;
int nbr;
int cconsole;
int mconsole;
struct timeval relax;
fd_set reading;
char *contpath;
char *filename;
int phase;
int proceed;
char buffer[1000];
count=0;
relaxtime=5;
status=0;
nbr=0;
cconsole=0;
mconsole=0;
relax.tv_sec=relaxtime;	/*sec relax time	*/
relax.tv_usec=0;
FD_ZERO(&reading);
contpath=getcontpath(contname);
(void) asprintf(&filename,"%s/%s",contpath,"console");
phase=0;
proceed=true;
while (proceed==true) {
  count++;
  switch (phase) {
    case 0	:	/*opening the M-Console*/
      if ((mconsole=open(filename,O_CREAT|O_WRONLY|O_TRUNC,S_IRUSR|S_IWUSR))<0) {
	(void) apl_alert(0,"%s, Unable to open mconsole <%s> (error=<%s>)",
			      OPEP,filename,strerror(errno));
	phase=999;
	status=-1;
	}
      break;
    case 1	:	/*opening the C-Console*/
      (void) free(filename);
      (void) asprintf(&filename,"%s/rootfs/dev/%s",contpath,"console");
      if ((cconsole=open(filename,O_RDWR|O_ASYNC))<0) {
	(void) apl_alert(0,"%s, Unable to open cconsole <%s> (error=<%s>)",
			      OPEP,filename,strerror(errno));
  	(void) close(mconsole);
	phase=999;
	status=-1;
        }
      break;
    case 2	:		/*waiting event		*/
      relax.tv_sec=relaxtime;	/*sec relax time	*/
      relax.tv_usec=0;
      FD_ZERO(&reading);
      FD_SET(cconsole,&reading);
      switch ((nbr=select(cconsole+1,&reading,(fd_set *)0,(fd_set *)0,&relax))) {
        case	-1    :
	  switch (errno) {
	    case EINTR:	/*received a signal, lets see...*/
	      break;	/*getting out fast		*/
	    default   :
	      (void) apl_alert(0,"%s, Container <%s> unexpected select "
			      	 "event (error=<%s>)",
				 OPEP,contname,strerror(errno));
	      break;
	    }
          break;
        case      0   :	/*timeout child lost ?	*/
	  switch (waitpid(cntpid,&status,WNOHANG)) {
	    case -1   :	/*exiting 		*/
	      (void) apl_alert(0,"%s, Container <%s> Unexpected error on "
			      	 "waitpid (error=<%s>)",
				 OPEP,contname,strerror(errno));
	      status=-1;
	      break;
	    case 0    :	/*process still up!	*/
	      if (relaxtime<RLX)
		relaxtime++;
	      phase--;	/*lets wait for console	*/
	      break;
	    default   :	/*process exited!	*/
	      (void) apl_alert(2,"Container <%s> exited with waitpid "
			         "status <%d -> %s>",
				 contname,status,strsignal(status));
	      if (WIFSIGNALED(status)==false) {
	        (void) apl_alert(2,"Container <%s> exit without signal",contname);
		status=0;
		}
	      else {
		status=WTERMSIG(status);
	        (void) apl_alert(1,"Container <%s> exited with signal <%d -> %s>",
				   contname,status,strsignal(status));
		}
	      break;	/*going to close phase	*/
	    }
          break;
	default	      :	/*data ready in console*/
	  if ((nbr=read(cconsole,buffer,sizeof(buffer)))>0) {
	    (void) write(mconsole,buffer,nbr);
	    relaxtime=1;
	    }		/*forwarded to mconsole	*/
	  phase--;	/*stay put on phase	*/
          break;
	}
      break;
    case 3	:	/*closing console	*/
      (void) close(mconsole);
      (void) close(cconsole);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(filename);
(void) free(contpath);
return status;

#undef	RLX
#undef	OPEP
}
